#include <stdint.h>
#include <grub/types.h>
#include <grub/dl.h>
#include <grub/command.h>
#include <grub/file.h>
#include <grub/env.h>
#include <grub/crypto.h>
#include "tpm.h"

GRUB_MOD_LICENSE("GPLv2+");

#define TPM_PCR_COUNT 24
#define SHA1_BYTES 20
#define SHA1_HEX_STR_SIZE (SHA1_BYTES * 2 + 1)

static int tpm_is_working;
static int tpm_file_pcr = -1;

static grub_size_t
bin2hex(char *hex_str, grub_size_t hex_str_size,
        const void *data, grub_size_t data_size)
{
  const uint8_t *octets = data;
  grub_size_t i, n;

  n = 0;
  for (i = 0; i < data_size; i++) {
    if (n + 2 < hex_str_size)
      grub_snprintf(hex_str + n, 3, "%02x", octets[i]);
    n += 2;
  }

  if (n < hex_str_size)
    hex_str[n] = 0;
  else if (hex_str_size > 0)
    hex_str[hex_str_size - 1] = 0;

  return n;
}

static void
grub_tpm_extend(int tpm_pcr, const void *data, grub_size_t size)
{
  void *fingerprint_context;
  uint8_t in_digest[SHA1_BYTES], out_digest[SHA1_BYTES];
  char in_digest_hex_str[SHA1_HEX_STR_SIZE], out_digest_hex_str[SHA1_HEX_STR_SIZE];

  if (!tpm_is_working || tpm_pcr < 0)
    return;

  fingerprint_context = grub_zalloc(GRUB_MD_SHA1->contextsize);
  if (!fingerprint_context) {
    grub_dprintf("tpm", "out of memory");
    return;
  }

  GRUB_MD_SHA1->init(fingerprint_context);
  GRUB_MD_SHA1->write(fingerprint_context, data, size);
  GRUB_MD_SHA1->final(fingerprint_context);

  grub_memcpy(in_digest, GRUB_MD_SHA1->read(fingerprint_context), sizeof(in_digest));
  grub_free(fingerprint_context);

  if (tpm_extend(tpm_pcr, in_digest, out_digest) != 0) {
    grub_dprintf("tpm", "failed to extend pcr#%d", tpm_pcr);
    return;
  }

  bin2hex(in_digest_hex_str, sizeof(in_digest_hex_str), in_digest, sizeof(in_digest));
  bin2hex(out_digest_hex_str, sizeof(out_digest_hex_str), out_digest, sizeof(out_digest));
  grub_dprintf("tpm", "extended pcr#%d: old=%s, new=%s\n", tpm_pcr, in_digest_hex_str, out_digest_hex_str);
}

static grub_file_t
grub_tpm_measure_file (grub_file_t io, const char *filename)
{
  int tpm_pcr = tpm_file_pcr;
  char *file_content;

  if (io->device->disk &&
      (io->device->disk->dev->id == GRUB_DISK_DEVICE_MEMDISK_ID
       || io->device->disk->dev->id == GRUB_DISK_DEVICE_PROCFS_ID))
    {
      grub_dprintf("tpm", "not measuring '%s'\n", filename);
      return io;
    }

  file_content = grub_malloc(io->size);
  if (!file_content)
    {
      return NULL;
    }
  if (grub_file_read (io, file_content, io->size) != (grub_ssize_t) io->size)
    {
      if (!grub_errno)
	grub_error (GRUB_ERR_FILE_READ_ERROR, N_("premature end of file %s"),
		    filename);
      grub_free (file_content);
      return NULL;
    }

  /*
   * Seek back to beginning of the file so that
   * next filter gets to see all of the file content
   */
  grub_file_seek (io, (grub_off_t) 0);

  grub_dprintf("tpm", "extending pcr#%d from file '%s'\n", tpm_pcr, filename);
  grub_tpm_extend(tpm_pcr, file_content, io->size);
  grub_free (file_content);

  return io;
}

static char *
grub_env_write_tpm_file_pcr(struct grub_env_var *var __attribute__((unused)),
			    const char *val)
{
  int tpm_pcr;
  char buf[32];

  if (!val || !*val) {
    tpm_file_pcr = -1;
    return NULL;
  }

  tpm_pcr = grub_strtol(val, 0, 0);
  if (grub_errno || tpm_pcr >= TPM_PCR_COUNT)
    return NULL;

  tpm_file_pcr = tpm_pcr;
  grub_snprintf(buf, sizeof(buf), "%d", tpm_pcr);
  return grub_strdup(buf);
}

static void
grub_dump_tpm_pcrs(void)
{
  unsigned tpm_pcr;
  uint8_t digest[SHA1_BYTES];
  char digest_hex_str[SHA1_HEX_STR_SIZE];

  for (tpm_pcr = 0; tpm_pcr < TPM_PCR_COUNT; tpm_pcr++) {
    if (tpm_pcr_read(tpm_pcr, digest, sizeof(digest)) == 0) {
      bin2hex(digest_hex_str, sizeof(digest_hex_str), digest, sizeof(digest));
      grub_printf("pcr#%02u: %s\n", tpm_pcr, digest_hex_str);
    } else
      grub_printf("pcr#%02u: digest unavailable\n", tpm_pcr);
  }
}

static grub_err_t
grub_cmd_tpm(grub_command_t cmd  __attribute__((unused)),
	     int argc, char **args)
{
  int tpm_pcr;

  if (!tpm_is_working)
    return grub_error(GRUB_ERR_IO, N_("TPM chip is not working"));

  if (argc > 0 && grub_strcmp (args[0], "extend") == 0) {
    if (argc < 3)
      return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("usage: tpm extend <pcr#> <data>"));

    tpm_pcr = grub_strtol(args[1], 0, 0);
    if (grub_errno || tpm_pcr < 0 || tpm_pcr >= TPM_PCR_COUNT)
      return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("invalid <pcr#>"));

    grub_dprintf("tpm", "extending pcr#%d with string '%s'\n", tpm_pcr, args[2]);
    grub_tpm_extend(tpm_pcr, args[2], grub_strlen(args[2]));
  } else if (argc > 0 && grub_strcmp (args[0], "pcrs") == 0) {
    grub_dump_tpm_pcrs();
  } else
    return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("usage: tpm {pcr|extend} ..."));

  return GRUB_ERR_NONE;
}

static grub_command_t cmd_tpm;

GRUB_MOD_INIT(tpm)
{
  tpm_is_working = tpm_init() == 0;

  grub_file_filter_register (GRUB_FILE_FILTER_TPM_MEASUREMENTS,
			     grub_tpm_measure_file);

  grub_register_variable_hook("tpm_file_pcr", NULL, grub_env_write_tpm_file_pcr);
  grub_env_export("tpm_file_pcr");

  cmd_tpm = grub_register_command("tpm", grub_cmd_tpm, 0,
				  N_("TPM related commands."));
}

GRUB_MOD_FINI(tpm)
{
  grub_unregister_command(cmd_tpm);
  grub_file_filter_unregister (GRUB_FILE_FILTER_TPM_MEASUREMENTS);
  grub_register_variable_hook("tpm_file_pcr", NULL, NULL);
}
