/*
 * Copyright (c) 2011 The Chromium OS Authors.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

/*
 * The code in this file is based on the article "Writing a TPM Device Driver"
 * published on http://ptgmedia.pearsoncmg.com.
 *
 * One principal difference is that in the simplest config the other than 0
 * TPM localities do not get mapped by some devices (for instance, by Infineon
 * slb9635), so this driver provides access to locality 0 only.
 */


#include <stdint.h>
#include <stdarg.h>
#include <grub/types.h>
#include <grub/mm.h>
#include <grub/misc.h>
#include <grub/time.h>
#include "tpm.h"

#define TPM_TIS_BASE_ADDRESS 0xFED40000

struct tpm_locality {
  uint32_t access;
  uint8_t padding0[4];
  uint32_t int_enable;
  uint8_t vector;
  uint8_t padding1[3];
  uint32_t int_status;
  uint32_t int_capability;
  uint32_t tpm_status;
  uint8_t padding2[8];
  uint8_t data;
  uint8_t padding3[3803];
  uint32_t did_vid;
  uint8_t rid;
  uint8_t padding4[251];
};

/*
 * This pointer refers to the TPM chip, 5 of its localities are mapped as an
 * array.
 */
#define TPM_TOTAL_LOCALITIES	5
static struct tpm_locality *lpc_tpm_dev =
(struct tpm_locality *)TPM_TIS_BASE_ADDRESS;

/* Some registers' bit field definitions */
#define TIS_STS_VALID                  (1 << 7) /* 0x80 */
#define TIS_STS_COMMAND_READY          (1 << 6) /* 0x40 */
#define TIS_STS_TPM_GO                 (1 << 5) /* 0x20 */
#define TIS_STS_DATA_AVAILABLE         (1 << 4) /* 0x10 */
#define TIS_STS_EXPECT                 (1 << 3) /* 0x08 */
#define TIS_STS_RESPONSE_RETRY         (1 << 1) /* 0x02 */

#define TIS_ACCESS_TPM_REG_VALID_STS   (1 << 7) /* 0x80 */
#define TIS_ACCESS_ACTIVE_LOCALITY     (1 << 5) /* 0x20 */
#define TIS_ACCESS_BEEN_SEIZED         (1 << 4) /* 0x10 */
#define TIS_ACCESS_SEIZE               (1 << 3) /* 0x08 */
#define TIS_ACCESS_PENDING_REQUEST     (1 << 2) /* 0x04 */
#define TIS_ACCESS_REQUEST_USE         (1 << 1) /* 0x02 */
#define TIS_ACCESS_TPM_ESTABLISHMENT   (1 << 0) /* 0x01 */

#define TIS_STS_BURST_COUNT_MASK       (0xffff)
#define TIS_STS_BURST_COUNT_SHIFT      (8)

/*
 * Error value returned if a tpm register does not enter the expected state
 * after continuous polling. No actual TPM register reading ever returns -1,
 * so this value is a safe error indication to be mixed with possible status
 * register values.
 */
#define TPM_TIMEOUT_ERR			(-1)

/* Error value returned on various TPM driver errors. */
#define TPM_DRIVER_ERR		(1)

/* 1 second is plenty for anything TPM does. */
#define MAX_DELAY_MS	1000

/* Retrieve burst count value out of the status register contents. */
static uint16_t burst_count(uint32_t status)
{
  return (status >> TIS_STS_BURST_COUNT_SHIFT) & TIS_STS_BURST_COUNT_MASK;
}

/*
 * Structures defined below allow creating descriptions of TPM vendor/device
 * ID information for run time discovery. The only device the system knows
 * about at this time is Infineon slb9635.
 */
struct device_name {
  uint16_t dev_id;
  const char * const dev_name;
};

struct vendor_name {
  uint16_t vendor_id;
  const char *vendor_name;
  const struct device_name *dev_names;
};

static const struct device_name infineon_devices[] = {
    {0xb, "SLB9635 TT 1.2"},
    {0, NULL}
};

static const struct vendor_name vendor_names[] = {
    {0x15d1, "Infineon", infineon_devices},
};

/*
 * Cached vendor/device ID pair to indicate that the device has been already
 * discovered.
 */
static uint32_t vendor_dev_id;

/* TPM access wrappers to support tracing */
static uint8_t tpm_read_byte(const uint8_t *ptr)
{
  uint8_t  ret = *(volatile uint8_t *)ptr;
  grub_dprintf("tpm_tis_lpc_io", "Read reg 0x%04x returns 0x%02x\n",
	       (uint32_t)(uintptr_t)ptr - (uint32_t)(uintptr_t)lpc_tpm_dev, ret);
  return ret;
}

static uint32_t tpm_read_word(const uint32_t *ptr)
{
  uint32_t  ret = *(volatile uint32_t *)ptr;
  grub_dprintf("tpm_tis_lpc_io", "Read reg 0x%04x returns 0x%08x\n",
	       (uint32_t)(uintptr_t)ptr - (uint32_t)(uintptr_t)lpc_tpm_dev, ret);
  return ret;
}

static void tpm_write_byte(uint8_t value, uint8_t *ptr)
{
  grub_dprintf("tpm_tis_lpc_io", "Write reg 0x%04x with 0x%02x\n",
	       (uint32_t)(uintptr_t)ptr - (uint32_t)(uintptr_t)lpc_tpm_dev, value);
  *(volatile uint8_t *)ptr = value;
}

static void tpm_write_word(uint32_t value, uint32_t *ptr)
{
  grub_dprintf("tpm_tis_lpc_io", "Write reg 0x%04x with 0x%08x\n",
	       (uint32_t)(uintptr_t)ptr - (uint32_t)(uintptr_t)lpc_tpm_dev, value);
  *(volatile uint32_t *)ptr = value;
}

/*
 * tis_wait_reg()
 *
 * Wait for at least a second for a register to change its state to match the
 * expected state. Normally the transition happens within microseconds.
 *
 * @reg - pointer to the TPM register
 * @mask - bitmask for the bitfield(s) to watch
 * @expected - value the field(s) are supposed to be set to
 *
 * Returns the register contents in case the expected value was found in the
 * appropriate register bits, or TPM_TIMEOUT_ERR on timeout.
 */
static int tis_wait_reg(uint32_t *reg, uint32_t *value_p,
			uint8_t mask, uint8_t expected)
{
  uint32_t time_ms = MAX_DELAY_MS;

  while (time_ms > 0) {
    uint32_t value = tpm_read_word(reg);
    if ((value & mask) == expected) {
      if (value_p)
	*value_p = value;
      return 0;
    }
    grub_millisleep(1); /* 1 ms */
    time_ms--;
  }
  return TPM_TIMEOUT_ERR;
}

/*
 * Probe the TPM device and try determining its manufacturer/device name.
 *
 * Returns 0 on success (the device is found or was found during an earlier
 * invocation) or TPM_DRIVER_ERR if the device is not found.
 */
int tis_init(void)
{
  uint32_t didvid = tpm_read_word(&lpc_tpm_dev[0].did_vid);
  unsigned i;
  const char *device_name = "unknown";
  const char *vendor_name = device_name;
  uint16_t vid, did;

  if (vendor_dev_id)
    return 0;  /* Already probed. */

  if (!didvid || (didvid == 0xffffffff)) {
    grub_dprintf("tpm_tis_lpc", "No TPM device found\n");
    return TPM_DRIVER_ERR;
  }

  vendor_dev_id = didvid;

  vid = didvid & 0xffff;
  did = (didvid >> 16) & 0xffff;
  for (i = 0; i < ARRAY_SIZE(vendor_names); i++) {
    int j = 0;
    uint16_t known_did;

    if (vid == vendor_names[i].vendor_id)
      vendor_name = vendor_names[i].vendor_name;

    while ((known_did = vendor_names[i].dev_names[j].dev_id) != 0) {
      if (known_did == did) {
	device_name =
	 vendor_names[i].dev_names[j].dev_name;
	break;
      }
      j++;
    }
    break;
  }

  grub_dprintf("tpm_tis_lpc", "Found TPM %s by %s\n", device_name, vendor_name);
  return 0;
}

/*
 * tis_senddata()
 *
 * send the passed in data to the TPM device.
 *
 * @data - address of the data to send, byte by byte
 * @len - length of the data to send
 *
 * Returns 0 on success, TPM_DRIVER_ERR on error (in case the device does
 * not accept the entire command).
 */
static uint32_t tis_senddata(const uint8_t * const data, uint32_t len)
{
  uint32_t offset = 0;
  uint16_t burst = 0;
  uint32_t max_cycles = 0;
  uint8_t locality = 0;
  uint32_t value;

  if (tis_wait_reg(&lpc_tpm_dev[locality].tpm_status, &value,
		   TIS_STS_COMMAND_READY, TIS_STS_COMMAND_READY) != 0) {
    grub_dprintf("tpm_tis_lpc", "failed to get 'command_ready' status\n");
    return TPM_DRIVER_ERR;
  }
  burst = burst_count(value);

  while (1) {
    unsigned count;

    /* Wait till the device is ready to accept more data. */
    while (!burst) {
      if (max_cycles++ == MAX_DELAY_MS) {
	grub_dprintf("tpm_tis_lpc", "failed to feed %d bytes of %d\n",
		     len - offset, len);
	return TPM_DRIVER_ERR;
      }
      grub_millisleep(1);
      burst = burst_count(tpm_read_word(&lpc_tpm_dev
					[locality].tpm_status));
    }

    max_cycles = 0;

    /*
     * Calculate number of bytes the TPM is ready to accept in one
     * shot.
     *
     * We want to send the last byte outside of the loop (hence
     * the -1 below) to make sure that the 'expected' status bit
     * changes to zero exactly after the last byte is fed into the
     * FIFO.
     */
    count = len - offset - 1;
    if (count > burst)
      count = burst;
    while (count--)
      tpm_write_byte(data[offset++],
		     &lpc_tpm_dev[locality].data);

    if (tis_wait_reg(&lpc_tpm_dev[locality].tpm_status, &value,
		     TIS_STS_VALID, TIS_STS_VALID) != 0 || !(value & TIS_STS_EXPECT)) {
      grub_dprintf("tpm_tis_lpc", "TPM command feed overflow\n");
      return TPM_DRIVER_ERR;
    }

    burst = burst_count(value);
    if ((offset == (len - 1)) && burst) {
      /*
       * We need to be able to send the last byte to the
       * device, so burst size must be nonzero before we
       * break out.
       */
      break;
    }
  }

  /* Send the last byte. */
  tpm_write_byte(data[offset++], &lpc_tpm_dev[locality].data);
  /*
   * Verify that TPM does not expect any more data as part of this
   * command.
   */
  if (tis_wait_reg(&lpc_tpm_dev[locality].tpm_status, &value,
		   TIS_STS_VALID, TIS_STS_VALID) != 0  || (value & TIS_STS_EXPECT)) {
    grub_dprintf("tpm_tis_lpc", "unexpected TPM status 0x%x\n", value);
    return TPM_DRIVER_ERR;
  }

  /* OK, sitting pretty, let's start the command execution. */
  tpm_write_word(TIS_STS_TPM_GO, &lpc_tpm_dev[locality].tpm_status);
  return 0;
}

/*
 * tis_readresponse()
 *
 * read the TPM device response after a command was issued.
 *
 * @buffer - address where to read the response, byte by byte.
 * @len - pointer to the size of buffer
 *
 * On success stores the number of received bytes to len and returns 0. On
 * errors (misformatted TPM data or synchronization problems) returns
 * TPM_DRIVER_ERR.
 */
static uint32_t tis_readresponse(uint8_t *buffer, uint32_t *len)
{
  uint16_t burst;
  uint32_t value;
  uint32_t offset = 0;
  uint8_t locality = 0;
  const uint32_t has_data = TIS_STS_DATA_AVAILABLE | TIS_STS_VALID;
  uint32_t expected_count = *len;
  int max_cycles = 0;

  /* Wait for the TPM to process the command. */
  if (tis_wait_reg(&lpc_tpm_dev[locality].tpm_status, &value,
		   has_data, has_data) != 0) {
    grub_dprintf("tpm_tis_lpc", "failed processing command\n");
    return TPM_DRIVER_ERR;
  }

  do {
    while ((burst = burst_count(value)) == 0) {
      if (max_cycles++ == MAX_DELAY_MS) {
	grub_dprintf("tpm_tis_lpc", "TPM stuck on read\n");
	return TPM_DRIVER_ERR;
      }
      grub_millisleep(1);
      value = tpm_read_word(&lpc_tpm_dev
			    [locality].tpm_status);
    }

    max_cycles = 0;

    while (burst-- && (offset < expected_count)) {
      buffer[offset++] = tpm_read_byte(&lpc_tpm_dev
				       [locality].data);

      if (offset == 6) {
	/*
	 * We got the first six bytes of the reply,
	 * let's figure out how many bytes to expect
	 * total - it is stored as a 4 byte number in
	 * network order, starting with offset 2 into
	 * the body of the reply.
	 */
	uint32_t real_length;
	grub_memcpy(&real_length,
		    buffer + 2,
		    sizeof(real_length));
	expected_count = grub_be_to_cpu32(real_length);

	if ((expected_count < offset) ||
	    (expected_count > *len)) {
	  grub_dprintf("tpm_tis_lpc", "bad response size %d\n", expected_count);
	  return TPM_DRIVER_ERR;
	}
      }
    }

    /* Wait for the next portion. */
    if (tis_wait_reg(&lpc_tpm_dev[locality].tpm_status, &value,
		     TIS_STS_VALID, TIS_STS_VALID) != 0) {
      grub_dprintf("tpm_tis_lpc", "failed to read response\n");
      return TPM_DRIVER_ERR;
    }

    if (offset == expected_count)
      break;	/* We got all we needed. */

  } while ((value & has_data) == has_data);

  /*
   * Make sure we indeed read all there was. The TIS_STS_VALID bit is
   * known to be set.
   */
  if (value & TIS_STS_DATA_AVAILABLE) {
    grub_dprintf("tpm_tis_lpc", "wrong receive status %x\n", value);
    return TPM_DRIVER_ERR;
  }

  /* Tell the TPM that we are done. */
  tpm_write_word(TIS_STS_COMMAND_READY, &lpc_tpm_dev
		 [locality].tpm_status);
  *len = offset;
  return 0;
}

int tis_open(void)
{
  uint8_t locality = 0; /* we use locality zero for everything. */

  if (tis_close())
    return TPM_DRIVER_ERR;

  /* now request access to locality. */
  tpm_write_word(TIS_ACCESS_REQUEST_USE, &lpc_tpm_dev[locality].access);

  /* did we get a lock? */
  if (tis_wait_reg(&lpc_tpm_dev[locality].access, NULL,
		   TIS_ACCESS_ACTIVE_LOCALITY,
		   TIS_ACCESS_ACTIVE_LOCALITY) != 0) {
    grub_dprintf("tpm_tis_lpc", "failed to lock locality %d\n", locality);
    return TPM_DRIVER_ERR;
  }

  tpm_write_word(TIS_STS_COMMAND_READY,
		 &lpc_tpm_dev[locality].tpm_status);
  return 0;
}

int tis_close(void)
{
  uint8_t locality = 0;

  if (tpm_read_word(&lpc_tpm_dev[locality].access) &
      TIS_ACCESS_ACTIVE_LOCALITY) {
    tpm_write_word(TIS_ACCESS_ACTIVE_LOCALITY,
		   &lpc_tpm_dev[locality].access);

    if (tis_wait_reg(&lpc_tpm_dev[locality].access, NULL,
		     TIS_ACCESS_ACTIVE_LOCALITY, 0) != 0) {
      grub_dprintf("tpm_tis_lpc", "failed to release locality %d\n", locality);
      return TPM_DRIVER_ERR;
    }
  }
  return 0;
}

int tis_sendrecv(const uint8_t *sendbuf, grub_size_t send_size,
		 uint8_t *recvbuf, grub_size_t *recv_len)
{
  if (tis_senddata(sendbuf, send_size)) {
    grub_dprintf("tpm_tis_lpc", "failed sending data to TPM\n");
    return TPM_DRIVER_ERR;
  }

  return tis_readresponse(recvbuf, (uint32_t *)recv_len);
}
